/***************************************************************************
 *   Copyright (C) 2004 by Virginia Tech                                   *
 *   ggrothau@vt.edu                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include<set>
#include<string>
#include "pqtree.h"

#include<fstream>
#include<cstdio>
#include<vector>
#include<map>
#include<queue>
#include<sstream>
#include<algorithm>

using namespace std;

vector<string> allGenes;
vector<string> allSamples;

struct bicluster
{
	set<int> genes;
	set<int> samples;
	vector<int> connections;
	string name;
	string color;
	bool operator<(const bicluster &a) const
	{	return name<a.name;	}
	bool operator==(const bicluster &a) const
	{	return name==a.name;	}

	//function cleans off the non-important stuff(connections,visited)
	//and switches the genes with the samples so that we can align in both directions
	void reverse()
	{
		visited=false;
		connections.clear();
		set<int> temp=genes;
		genes=samples;
		samples=temp;
	}

	
	bool visited;
};

int atoI(string in)
{
	stringstream ss;
	ss<<in;
	int out;
	ss>>out;
	return out;
}

//trims whitespace from both sides of a string
//to aid in reading in data consistently
string trim(string in)
{
	string out;
	bool start=false;
	for(int i=0;i<in.size();i++)
	{
		if(!(in[i]==' ' || in[i]=='\t'))
			start=true;
		if(start)
			out+=in[i];
	}
	reverse(out.begin(),out.end());
	in=out;
	out="";
	start=false;
	for(int i=0;i<in.size();i++)
	{
		if(!(in[i]==' ' || in[i]=='\t'))
			start=true;
		if(start)
			out+=in[i];
	}
	reverse(out.begin(),out.end());
	return out;
}
		

vector<bicluster> input(string infile)
{
	ifstream in(infile.c_str());	
	
	map<string,int> geneIdx;
	map<string,int> sampleIdx;
	allSamples.push_back("");
	allGenes.push_back("");
	vector<bicluster> out;
	
	int numsamples, numgenes;
	//this gets the first cluster name and integers describing it - number of samples and number of genes
	string name,buf;
	getline(in,name,'_');
	
	getline(in,buf,'_');
	numsamples=atoI(buf);
	
	getline(in,buf,'\t');
	numgenes=atoI(buf);
	
	while(!in.fail())
	{
		bicluster c;
		c.name=name;
		
		//ignore the token, it seems to be a non-included color
		getline(in,buf,'\t');
		c.color=buf;

		//read in the samples
		for(int i=0;i<numsamples;i++)
		{
			getline(in,buf,'\t');
			string bf=trim(buf);
			if(sampleIdx[bf]==0)
			{
				sampleIdx[bf]=allSamples.size();
				allSamples.push_back(bf);
			}
			c.samples.insert(sampleIdx[bf]);
		}
		//ignore extra tab separator
		//getline(in,buf,'\t');
		
		//read in the genes
		for(int i=0;i<numgenes-1;i++)
		{
			getline(in,buf,'\t');
			string bf=trim(buf);
			if(geneIdx[bf]==0)
			{
				geneIdx[bf]=allGenes.size();
				allGenes.push_back(bf);
			}
			c.genes.insert(geneIdx[bf]);
		}
		//get last gene terminated by endline instead of a tab
		getline(in,buf,'\n');
		string bf=trim(buf);
		if(geneIdx[bf]==0)
		{
			geneIdx[bf]=allGenes.size();
			allGenes.push_back(bf);
		}
		c.genes.insert(geneIdx[bf]);
		
		//add the cluster to our list
		out.push_back(c);
		//get the next line (or EOF)
	
		getline(in,name,'_');
	
		getline(in,buf,'_');
		numsamples=atoI(buf);
	
		getline(in,buf,'\t');
		numgenes=atoI(buf);
	}
	return out;
}

//builds a undirected graph where clusters are connected if they share a gene
vector<bicluster> buildggraph(vector<bicluster> in)
{
	for(int i=0;i<in.size();i++)
	{
		for(int j=i+1;j<in.size();j++)
		{
			set<int> temp;
			set<int>::iterator end;
			set_intersection(in[i].genes.begin(), in[i].genes.end(),in[j].genes.begin(), in[j].genes.end(),
					inserter(temp,temp.begin()) );
			
			if(temp.size()>0)
			{
				in[i].connections.push_back(j);
				in[j].connections.push_back(i);
			}
		}
	}
	return in;
}

//connected components algorithm (undirected)
//returns a 2d vector of vectors - the subvectors are a list of connected components
vector<vector<int> > connected(vector<bicluster> in)
{
	vector<vector<int> > out;
	
	for(int i=0;i<in.size();i++)
		in[i].visited=false;

	//connected components, flood fill
	for(int i=0;i<in.size();i++)
	{
		if(!in[i].visited)
		{
			queue<int> q;
			q.push(i);
			vector<int> v;
			while(!q.empty())
			{
				int cur=q.front();
				q.pop();
				if(!in[cur].visited)
				{
					for(int i=0;i<in[cur].connections.size();i++)
						q.push(in[cur].connections[i]);
					v.push_back(cur);
				}

				in[cur].visited=true;
			}
			out.push_back(v);
		}
	}
	return out;
			
}

//vector<pqtree> addRepeats(vector<bicluster> clusters,vector<alignmentList> lists);

//mergesorter is for sorting merge scores in the doMerge function
struct mergeSorter{
	double score;
	int x,y;
	bool operator==(const mergeSorter& a) const
	{	return a.score==score;	}
	bool operator<(const mergeSorter& a) const
	{	return a.score<score;	}
};

vector<pqtree<int>*> doMerge(vector<bicluster> in)
{
	
	mergeSorter m;

	
	//build initial vector of pqtrees
	vector<pqtree<int>*> v;
	set<int> allgenes;
	
	for(int i=0;i<in.size();i++)
		allgenes=setunion(allgenes,in[i].genes);
			
	for(int i=0;i<in.size();i++)
	{
		pqtree<int> *p=new pqtree<int>(allgenes);
		p->reduce(in[i].genes);
		v.push_back(p);
	}
	
	//no need to keep this memory laying around for the whole process
	allgenes.clear();
	
	//loop until nothing can merge anymore
	bool change=true;
	while(change && v.size()>1)
	{
          //this statement is for estimating time remaining, can be removed
//          cout<<v.size()<<endl;
		
		change=false;
		double max=0;
		int maxi=0,maxj=0;
		
		vector<mergeSorter> vms;
		
		//loop through all pairs of two alignment lists to find the
		//best available alignment.  Can change scoring metric here
		//currently it is just the largest number of elements that 
		//would intersect in the two alignments
		for(int i=0;i<v.size();i++)
		{
			pqtree<int> *p=v[i];
			for(int j=i+1;j<v.size();j++)
			{		
				pqtree<int> *q=v[j];
				//pqtree<int> copy(*q);
				
				double score=-1;
				
				//list<set<int> > ll=p->getReductions();
				//if we can merge these trees by reducing one with the other
				//if(copy.reduceAll(p->getReductions()))
				//{
					
				//our score metric is the intersection/union of the contained values
				score =double(setintersection(p->getContained(),q->getContained()).size())/double(setunion(p->getContained(),q->getContained()).size());	
				//}
			
				m.score=score;
				m.x=i;m.y=j;
				//if(score>0)
					vms.push_back(m);
			}
		}
		
		//we found a doable alignment
		sort(vms.begin(),vms.end());
		pqtree<int> *copy;
		int vmsi;
				
		for(vmsi=0;vmsi<vms.size();vmsi++)
		{
			m=vms[vmsi];
			copy = new pqtree<int>(*v[m.y]);

			//TODO: next line is breaking something
			if( !(copy->reduceAll(v[m.x]->getReductions())) )
				delete copy;
			else
				break;
		}
		
		if(vmsi<vms.size())
		{
			change=true;
			
			//erase the old alignment lists
			delete v[vms[vmsi].x];
			delete v[vms[vmsi].y];
			v.erase(v.begin()+vms[vmsi].y,v.begin()+vms[vmsi].y+1);
			v.erase(v.begin()+vms[vmsi].x,v.begin()+vms[vmsi].x+1);
			//put the new merged list back in our vector
			v.push_back(copy);
		}
		
	}
	//cout<<v.size()<<endl;
	
	return v;
}

/*
vector<alignmentList> addRepeats(vector<bicluster> clusters,vector<alignmentList> lists)
{
	//for each alignment list (1 cluster), attempt to align it to a list it isn't already aligned to
	for(int i=0;i<clusters.size();i++)
	{
		bicluster cur=clusters[i];
		string basename=cur.name;

		for(int j=0;j<lists.size();j++)
		{
			//no point in aligning the same list twice
			if(lists[j].containsCluster(cur))
				continue;
			lists[j].overlapMode();

			alignmentList p(cur);
			p.overlapMode();
			//can we align them exactly without additional size?
			if(lists[j]*p>0)
			{
				alignmentList temp=lists[j]+p;	//merge, so we can show repeated regions
				if(temp.getOrder().size()==lists[j].getOrder().size())
					lists[j]=temp;
				//c++;	//increment the itemset name
			}
		}
	}
	return lists;
}*/

//prints the results of this algorithm to standard out
//this output format isn't really for human consumption
//more so it is useful for another software component (drawlayout)
bool openedonce=false;
void printresults(vector<pqtree<int>*> in,string outfile,vector<bicluster> data)
{
	ofstream out;
	if(outfile!="stdout")
		if(openedonce==false)
			out.open(outfile.c_str());
		else
			out.open(outfile.c_str(), ofstream::out | ofstream::app);
	openedonce=true;
	
	//print the genes in the order they should be displayed
	//one per line
	vector<int> layout;
	map<pqtree<int>*,int> index;
	for(int i=0;i<in.size();i++)
	{
		index[in[i]]=layout.size();
		list<int> order=in[i]->reducedFrontier();
		
		for(list<int>::iterator j=order.begin();j!=order.end();j++)
		{
			if(outfile=="stdout")
				cout<<allGenes[*j]<<endl;
			else
				out<<allGenes[*j]<<endl;
			layout.push_back(*j);
		}
	}
	//skip a line
	if(outfile=="stdout")
		cout<<endl;
	else
		out<<endl;
	//print each cluster<tab>color<tab>number of genes it contains<tab>index of first gene where it appears
	for(int i=0;i<in.size();i++)
	{
		list<set<int> > v=in[i]->getReductions();
		map<int,bool> seen;
		for(int k=0;k<data.size();k++)
			seen[k]=false;
		for(list<set<int> >::iterator j=v.begin();j!=v.end();j++)
		{
			//figure out which bicluster we have
			bicluster b;
			
			for(int k=0;k<data.size();k++)
				if(*j==data[k].genes  && seen[k]==false)
				{
					seen[k]=true;
					b=data[k];
					break;
				}
			
			//the index of the first element in the bicluster
			//within the frontier of b
			int frontierIndex=0;
			list<int> f=in[i]->reducedFrontier();
			int k2=0;
			for(list<int>::iterator k=f.begin();k!=f.end();k++)
			{
				if(setfind(b.genes,*k))
				{
					frontierIndex=k2;
					break;
				}
				k2++;
			}
			if(outfile=="stdout")
				cout<<b.name<<"\t"<<b.color<<"\t"<<b.genes.size()<<"\t"<<index[in[i]]+frontierIndex<<endl;
			else
				out<<b.name<<"\t"<<b.color<<"\t"<<b.genes.size()<<"\t"<<index[in[i]]+frontierIndex<<endl;
				
		}
	}
	//skip a line 
	if(outfile=="stdout")
		cout<<endl;
	else
		out<<endl;
	if(outfile!="stdout")
		out.close();
}

void testbed()
{
	pqtree<int> *P;
	set<int> S;
	for(int i=1;i<9;i++)
		S.insert(i);
	P=new pqtree<int>(S);
	
	P->print();
	
	S.clear();
	S.insert(4);
	S.insert(3);
	P->reduce(S);
	P->print();
	
	S.clear();
	S.insert(6);
	S.insert(4);
	S.insert(3);
	P->reduce(S);
	P->print();
	
	S.clear();
	S.insert(4);
	S.insert(3);
	S.insert(5);
	P->reduce(S);
	P->print();
		
	S.clear();
	S.insert(4);
	S.insert(5);
	P->reduce(S);
	P->print();/*
	
	
	S.clear();
	S.insert(2);
	S.insert(6);
	P->reduce(S);
	P->print();
	
	S.clear();
	S.insert(1);
	S.insert(2);
	P->reduce(S);
	P->print();
	
	S.clear();
	S.insert(4);
	S.insert(5);
	P->reduce(S);
	P->print();
	
/*	S.clear();
	S.insert(2);
	S.insert(3);
	P->reduce(S);
	P->print();
	
	
	S.clear();
	S.insert(5);
	S.insert(6);cerr<<"*"<<endl;
	P->reduce(S);cerr<<"*"<<endl;
	P->print();*/
}
	

int main(int argc, char **argv)
{
	if(false)
		testbed();
	else
	{
	string inputfile="";
	string outputfile="stdout";
	for(int i=1;i<argc;i++)
	{	
		//output file parameter
		if(argv[i]==string("-o"))
			outputfile=argv[++i];
		else if(argv[i]==string("--help") || argv[i]==string("-help") || argv[i]==string("-h"))
		{
			cout<<"Bicluster Layout Tool"<<endl;
			cout<<"Usage: layout [<OPTION>] <inputfile>"<<endl;
			cout<<" -o <file>	Specify Output File"<<endl;
			cout<<" -h		This Screen"<<endl;
			return 0;
		}
		//default parameter (bicluster input file)
		else if(argv[i-1][0]!='-' && argv[i][0]!='-')
			inputfile=argv[i];
		//handles undefined flags
		else
		{
			cerr<<argv[i]<<" is unrecognized.\n";
			return 1;
		}
	}
	if(inputfile=="")
	{
		cerr<<"Error, no input file.\n -h for Help\n";
		exit(1);
	}
	vector<bicluster> data;
	vector<pqtree<int>*> results;
	
	//read in data
	data=input(inputfile);
	
	data=buildggraph(data);

	vector<vector<int> > cc;
	cc=connected(data);
	
	cout<<"Gene Layout Calculations: ";
	cout<<cc.size()<<"\tProcessing connected component "<<endl;
	//process each component seperately
	for(int i=0;i<cc.size();i++)
	{
//		cout<<"Component "<<i+1<<":"<<endl;
		cout<<" "<<i+1;
		vector<bicluster> component;
		for(int j=0;j<cc[i].size();j++)
			component.push_back(data[cc[i][j]]);

		vector<pqtree<int>*> run=doMerge(component);
		for(int j=0;j<run.size();j++)
			results.push_back(run[j]);		
		
	}
        cout << endl;
	printresults(results,outputfile,data);
	
	//reverse the genes/samples to run algorithm again
	vector<string> temp;
	temp=allGenes;
	allGenes=allSamples;
	allSamples=temp;
	for(int i=0;i<data.size();i++)
		data[i].reverse();

	//clear out other data structures
	results.clear();cc.clear();
	
	//Do it again!
	data=buildggraph(data);
	cc=connected(data);
	
	cout<<"Sample Layout Calculations: ";
	cout<<cc.size()<<"\tProcessing connected component"<<endl;
	//process each component seperately
	for(int i=0;i<cc.size();i++)
	{
		cout<<" "<<i+1;
		vector<bicluster> component;
		for(int j=0;j<cc[i].size();j++)
			component.push_back(data[cc[i][j]]);
		vector<pqtree<int>*> run=doMerge(component);
		for(int j=0;j<run.size();j++)
			results.push_back(run[j]);		
	}
	printresults(results,outputfile,data);
        cout << endl;
	cout<<"Done."<<endl;
	}
}


